/*! \file  ControlsKeyboard.cpp
 *  \brief ControlsKeyboard class definition
 */

#include "ControlsKeyboard.h"

// Define our default key translation table
#include "ControlsKeyboard_defaults.cpp"

ControlsKeyboard::ControlsKeyboard (Camera *cam, Stage *stg, Collision *col, Networkh *net)
{
	this->cam = cam;
	this->net = net;
	this->stg = stg;
	this->col = col;
	this->msg = stg->getMessageControl();
	this->console = new Console(net, msg);

	/* set our default configuration */
	configured_keys = new _key_t[NKEYS];
	memcpy(configured_keys, default_keys, sizeof(_key_t)*NKEYS);
	if ( loadConfig(KBD_CONF, configured_keys) < 0 )
	{
		fprintf(stderr,"[E] Invalid keyboard config file.\n");
	}

	running=false;

	animation=0;
	nextShoot=0;
	jump=0;
	gojump=false;
	jumpTime=SDL_GetTicks();

	speed=0.0008f;
	turn_speed=0.04f;

	weaponStyle=FACE2FACE;

	k=0;
}

ControlsKeyboard::~ControlsKeyboard()
{
	delete[] configured_keys;
	delete console;
}

// XXX: should not setAnimation from Controls!
void ControlsKeyboard::input(bool *bye)
{
	Movement movement;       // our movement
	Orientation orientation; // our head/camera orientation
	this->quitGame = bye;

	moving = false;
	gojump = false;

	keys = SDL_GetKeyState(NULL);
	SDL_PollEvent(&ev);

	/* NORMAL MODE */
	if (!msg->isTalkMessageActive()) 
	{
		FKey fkey;
		if (keys[SDLK_x] == SDL_PRESSED)
		{
			GL3f pos = cam->getPos();
			printf("POS:%f %f %f\n",pos.x,pos.y,pos.z);
		}

		if (keys[SDLK_l] == SDL_PRESSED)
		{
			DRAW_TRACES^=true;
			keys[SDLK_l]=0; // ?
		}

		for (int i=0; i < NKEYS; i++)
		{
			fkey=(FKey)i;
			(this->*configured_keys[(FKey)fkey].hook)
			(getIsFKeyPressed(fkey), &movement, &orientation);
		}
	}
	else
	{
	/* INPUT MESSAGE MODE */
		if (keys[SDLK_ESCAPE]==SDL_PRESSED) //getIsFKeyPressed(CANCEL)) 
		{
			if (ev.type == SDL_KEYDOWN)
			{
			msg->destroyTalkMessage();
			//setFKeyState(CANCEL, KEY_RELEASED);
			keys[SDLK_ESCAPE]=SDL_RELEASED;
			}
		}
		else if (keys[SDLK_RETURN]==SDL_PRESSED)
		{
			console->parseCommand(msg->getMessage());
			msg->destroyTalkMessage();
			keys[SDLK_RETURN]=SDL_RELEASED;
			//setFKeyState(ACCEPT,KEY_RELEASED);
		}
		else {
			//if (getIsFKeyPressed(DELETE)) 
			if (keys[SDLK_ESCAPE]==SDL_PRESSED)
			{
				if (ev.type == SDL_KEYDOWN)
				{
				//if (!getWasFKeyPressed(DELETE))
				msg->delFromTalkMessage(1);
			//	setFKeyState(DELETE, KEY_PRESSED);
				keys[SDLK_ESCAPE]=SDL_RELEASED;
				return;
				}
			}
			//else setFKeyState(DELETE, KEY_RELEASED);

			char *ckey;
			char str[128];
			for (int i=SDLK_FIRST; i<SDLK_LAST; i++)
			{
				if (keys[i] == SDL_PRESSED)
				{
					keys[i]=SDL_RELEASED;
					ckey = SDL_GetKeyName((SDLKey)i);
					str[0]='\0';

					if (ckey)
					{
					strcpy(str,ckey);
						if (ckey[1]!='\0')
						{
						// TODO: check for "shift", "control"
							if (!strcmp(ckey,"space"))
							{
								strcpy(str," ");
							} else
							if (!strcmp(ckey,"backspace"))
							{
								msg->delFromTalkMessage(1); return;
							}
						}
						for (int j=0; str[j] != '\0'; j++)
						{
							msg->addToTalkMessage(str[j]);
						}
					}
				}
			}
		}
	}

	if (moving)
	{
		stepVar+=0.1;
		step=0.3f*sin(stepVar);
	}else{
		step=0.0f;
		setAnimation(NOT_MOVING,2);
	}

	stg->setWalkEffect(step);

	/* gravetat */
	if (jumpTime < SDL_GetTicks())
	{
		movement.pos.y-=0.02f;
	}

	/* check collisions */
	bool jam = col->checkMovement(&orientation,&movement);

	/* JUMP */
	if (gojump && !jam)
	{
		jump=0;
	}
	if ( jumpTime < SDL_GetTicks() )
	{
		jumpTime = SDL_GetTicks()+10;

		if (jump > 0)
		{
		movement.pos.y+=(0.004*(jump));
		jump--;
		}
	}

	/* change position/orientation */
	cam->addOrientation(&orientation);
	cam->addMovement(&movement);
}

bool ControlsKeyboard::getIsFKeyPressed (FKey key)
{
	return keys[configured_keys[key].sdl]==SDL_PRESSED ;
}

bool ControlsKeyboard::getIsFKeyDown ()
{
	return (ev.type==SDL_KEYDOWN);
}

bool ControlsKeyboard::getWasFKeyPressed (FKey key)
{
	return configured_keys[key].pressed==KEY_PRESSED;
}

void ControlsKeyboard::setFKeyState (FKey key, KeyState state)
{
	configured_keys[key].pressed=state;
}

void ControlsKeyboard::f_walk_f (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
	{
		moving=true;
		walk((float)renderDelay, movement);
	}
}

void ControlsKeyboard::f_walk_b (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
	{
		moving=true;
		walk(-(float)renderDelay, movement);
	}
}

void ControlsKeyboard::f_run (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
		movement->run=true;
}

void ControlsKeyboard::f_turn_l (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
	{
		moving=true;
		turn(-(float)renderDelay, orientation);
	}
}		

void ControlsKeyboard::f_turn_r (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
	{
		moving=true;
		turn((float)renderDelay, orientation);
	}
}

void ControlsKeyboard::f_strafe_l (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
	{
		moving=true;
		strafe(-8, movement);
	}
}

void ControlsKeyboard::f_strafe_r (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
	{
		moving=true;
		strafe (8, movement);
	} 
}

void ControlsKeyboard::f_jump (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
	{
		if (jump == 0)
		{
		gojump=true;
		jump = 14;
		}
	}
}

void ControlsKeyboard::f_center_v (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
		cam->centerV();
}

void ControlsKeyboard::f_look_up (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
		orientation->addV((M_PI/90.0)*turn_speed*4);
}

void ControlsKeyboard::f_look_down (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
		orientation->addV(-(M_PI/90.0)*turn_speed*4);
}

void ControlsKeyboard::f_fire (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
	{
		if (SDL_GetTicks() < nextShoot) return;

		nextShoot=SDL_GetTicks()+stg->getCamaraWeapon()->getDelay();
		setAnimation(SHOOTING,1);
		GL3f pos(stg->getCamaraPos().x,stg->getCamaraPos().y+PLAYER_HEIGHT,stg->getCamaraPos().z);
		GL3f vec((float)cos(yaw)*(float)cos(yawV),(float)sin(yawV),(float)sin(yaw)*(float)cos(yawV));
		fprintf(stderr,"SHOOTING"); // OUTPUT::DEBUG

		if (NETWORK)
			net->cli_sendShot(pos.x,pos.y,pos.z,vec.x,vec.y,vec.z);
	}
	else setAnimation(NOT_SHOOTING,1);
}

void ControlsKeyboard::f_chweap (bool active, Movement *movement, Orientation *orientation)
{
// TODO : change weapon
#if 0
	if (active) {
		if (!getWasFKeyPressed(CHWEAP)) {
			msg->addMessage("Weapon Changed!!!");
			net->cli_sendWeap(-1);
			setFKeyState(CHWEAP, KEY_PRESSED);
		}
	}
	else setFKeyState(CHWEAP, KEY_RELEASED);
#endif
}

void ControlsKeyboard::f_map (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
	{
		if (!getWasFKeyPressed(MAP))
		{
			stg->swapViewMap();
			setFKeyState(MAP, KEY_PRESSED);
		}
	}
	else setFKeyState(MAP, KEY_RELEASED);
}

void ControlsKeyboard::f_talk (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
	{
		if (!getWasFKeyPressed(TALK))
		{
			msg->createTalkMessage();
			setFKeyState(TALK, KEY_PRESSED);
		}
	}
	else setFKeyState(TALK, KEY_RELEASED);
}

void ControlsKeyboard::f_screenshot (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
	{
		if (!getWasFKeyPressed(SCREENSHOT))
		{
			saveScreenShot();
			setFKeyState(SCREENSHOT, KEY_PRESSED);
		}
	}
	else setFKeyState(SCREENSHOT, KEY_RELEASED);

}

void ControlsKeyboard::f_pause (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
		fprintf(stderr,"game paused\n"); //fprintf(stderr, "game paused\n");
}

void ControlsKeyboard::f_quit (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
	{
		*quitGame = true;
	}

}

void ControlsKeyboard::f_fovp (bool active, Movement *movement, Orientation *orientation)
{
#warning is/was FKeyActive?
#warning Take FOV from a class?
	if (active)
		if (FOV < 360) FOV+=1;
}

void ControlsKeyboard::f_fovm (bool active, Movement *movement, Orientation *orientation)
{
#warning is/was FKeyActive?
#warning Take FOV from a class?
	if (active)
		if (FOV > 0) FOV-=1;
}

void ControlsKeyboard::f_showpos (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
	{
		if (!getWasFKeyPressed(SHOWPOS))
		{
			GL3f pos=cam->getPos();
			//fprintf(stderr, "position: %f %f %f",pos.x,pos.y,pos.z);
			setFKeyState(SHOWPOS, KEY_PRESSED);
		}
	}
	else setFKeyState(SHOWPOS, KEY_RELEASED);
}

void ControlsKeyboard::f_showitems (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
	{
		if (!getWasFKeyPressed(SHOWITEMS))
		{
			(stg->getItem(0)->isVisible())
				? stg->getItem(0)->setExplosion(true)
				: stg->getItem(0)->setExplosion(false);
			setFKeyState(SHOWITEMS, KEY_PRESSED);
		}
	}
	else setFKeyState(SHOWITEMS, KEY_RELEASED);

}

void ControlsKeyboard::f_showthirdp (bool active, Movement *movement, Orientation *orientation)
{
#warning Stage::setThirdPerson vs Camera::setThirdView
	if (active)
	{
		if (!getWasFKeyPressed(SHOWTHIRDP))
		{
			cam->swapThirdView();
			stg->setThirdPerson(cam->getThirdView());
			setFKeyState(SHOWTHIRDP, KEY_PRESSED);
		}
	}
	else setFKeyState(SHOWTHIRDP, KEY_RELEASED);

}

void ControlsKeyboard::f_kill (bool active, Movement *movement, Orientation *orientation)
{
#if 0
	if (active) {
		if (!getWasFKeyPressed(KILL)) {
			(stg->getEnemy(0)->isDying())?stg->getEnemy(0)->reviveEnemy():stg->getEnemy(0)->killEnemy();
			setFKeyState(KILL, KEY_PRESSED);
		}
	}
	else setFKeyState(KILL, KEY_RELEASED);
#endif
}

#warning Using static varibles on f_mode_*? (PORTALES, FOG, BOUNDING...)
void ControlsKeyboard::f_mode_portal (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
	{
		if (!getWasFKeyPressed(MODE_PORTAL))
		{
			PORTALES^=1;
			if (PORTALES) msg->addMessage("Portal mode on");
			else msg->addMessage("Portal mode off");
			setFKeyState(MODE_PORTAL, KEY_PRESSED);
		}
	}
	else setFKeyState(MODE_PORTAL, KEY_RELEASED);
}

void ControlsKeyboard::f_mode_bounding (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
	{
		if (!getWasFKeyPressed(MODE_BOUNDING))
		{
			BOUNDING?BOUNDING=false:BOUNDING=true;
			if (BOUNDING) msg->addMessage("Bounding mode on");
			else msg->addMessage("Bounding mode off");
			setFKeyState(MODE_BOUNDING, KEY_PRESSED);
		}
	}
	else setFKeyState(MODE_BOUNDING, KEY_RELEASED);
}

void ControlsKeyboard::f_mode_painting (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
	{
		if (!getWasFKeyPressed(MODE_PAINTING))
		{
			if (!DEBUG) msg->addMessage("Painting triangles");
			else if (DEBUG==1) msg->addMessage("Painting everything");
			else msg->addMessage("Painting normal");
			if (DEBUG<2) DEBUG++;
			else DEBUG=0;
			setFKeyState(MODE_PAINTING, KEY_PRESSED);
		}
	}
	else setFKeyState(MODE_PAINTING, KEY_RELEASED);
}

void ControlsKeyboard::f_mode_textures (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
	{
		if (!getWasFKeyPressed(MODE_TEXTURES))
		{
			TEXTURA^=1;
			if (TEXTURA) msg->addMessage("Painting Textures");
			else msg->addMessage("Tripi mode on :)");
			setFKeyState(MODE_TEXTURES, KEY_PRESSED);
		}
	}
	else setFKeyState(MODE_TEXTURES, KEY_RELEASED);
}

void ControlsKeyboard::f_mode_fog (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
	{
		if (!getWasFKeyPressed(MODE_FOG))
		{
			FOG^=1;
			if (FOG)
				msg->addMessage("FOG ENABLED");
			else
				msg->addMessage("FOG DISABLED");
			setFKeyState(MODE_FOG,  KEY_PRESSED);
		}
	}
	else setFKeyState(MODE_FOG, KEY_RELEASED);
}

void ControlsKeyboard::f_mode_light (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
	{
		if (!getWasFKeyPressed(MODE_LIGHT))
		{
			LIGHT^=1;
			if (LIGHT)
				msg->addMessage("LIGHTS ENABLED");
			else
				msg->addMessage("LIGHTS DISABLED");

			setFKeyState(MODE_LIGHT,  KEY_PRESSED);
		}
	}
	else setFKeyState(MODE_LIGHT, KEY_RELEASED);
}

void ControlsKeyboard::f_mode_col (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
	{
		if (!getWasFKeyPressed(MODE_COL))
		{
			DEBUG_COLLISION^=1;
			if (DEBUG_COLLISION)
				msg->addMessage("DEBUG COLLISIONS ON");
			else
				msg->addMessage("DEBUG COLLISIONS OFF");
			setFKeyState(MODE_COL,  KEY_PRESSED);
		}
	}
	else setFKeyState(MODE_COL, KEY_RELEASED);
}

void ControlsKeyboard::f_fly_up (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
	{
		movement->pos.y=+0.006;
	}
}

void ControlsKeyboard::f_fly_down (bool active, Movement *movement, Orientation *orientation)
{
	if (active)
	{
		movement->pos.y=-0.006;
	}
}

void ControlsKeyboard::walk (float dir, Movement *movement)
{
#warning Fusion with strafe?
	float spd = speed;

	if (running) spd*=2;

#if 0
	GL3f _posCam;

	//if (!NETWORK) // XXX ?
	//	_posCam=stg->getCamaraPos();



	_posCam.x+=(dir*((float)cos(yaw)))*spd;
	_posCam.z+=(dir*((float)sin(yaw)))*spd;
#endif

	if (dir>0) 
		setAnimation(RUNNING_FORWARD,2);
	else 
		setAnimation(RUNNING_BACKWARD,2);

	float yaw = cam->getRotation(); //cam->getYaw().x;

	movement->pos.x+=(dir*((float)cos(yaw)))*spd;
	movement->pos.y+=0;
	movement->pos.z+=(dir*((float)sin(yaw)))*spd;

	//stg->setCamaraObj(yaw,yawV);

	//if (!NETWORK)
	//	stg->setCamaraPos(_posCam);

}

void ControlsKeyboard::turn (float dir, Orientation *orientation)
{
#warning Fusion with look_up/down?
#if 0
	GL3f _posCam;
	if (!NETWORK) // XXX X?
		_posCam=stg->getCamaraPos();

	if (yaw>=2*M_PI) yaw-=2*M_PI;
	if (yaw<=0) yaw+=2*M_PI;
	yaw=yaw+dir*(M_PI/90.0)*turn_speed;
#endif
	orientation->setH(dir*(M_PI/90.0f)*turn_speed);
	orientation->setV(0);
}

void ControlsKeyboard::saveScreenShot ()
{
	SDL_Surface *shot =
		SDL_CreateRGBSurface(SDL_SWSURFACE,WIDTH,HEIGHT,24,
#if SDL_BYTEORDER == SDL_LIL_ENDIAN /* OpenGL RGBA masks */
				0x0000FF, 0x00FF00, 0xFF0000, 0x000000
#else
				0xFF0000, 0x00FF00, 0x0000FF, 0x000000
#endif
				);

	glReadPixels(0,0,WIDTH,HEIGHT,GL_RGB,GL_UNSIGNED_BYTE,shot->pixels);
#if WIN32
	SDL_SaveBMP(shot,"screenshot.bmp");
#else
	SDL_SaveBMP(shot,"/tmp/screenshot.bmp");
#endif
	SDL_FreeSurface(shot);

	msg->addMessage("Screenshot saved to 'screenshot.bmp'");
}

/*!
 * \return Returns -1 if an error occurred, 0 if could not find the config
 *         file and 1 if the config file was succesfully loaded.
 * \sa saveConfig()
 */
int ControlsKeyboard::loadConfig (char *filename, _key_t *keys)
{
	register int i;
	char str[1024];
	FILE *file;
	bool fine;

	chCfgDir();

	file = fopen(filename, "r");
	if (!file)
		return 0;

	while (!feof(file))
	{
		fscanf(file, "%s", str);
		if (feof(file)) break;

		for (i=0; i<NKEYS; i++)
		{
			if (!strcmp(str, keys[i].name))
			{
				fscanf(file, "%s", str);
				if (feof(file)) return 0;
				for (int j=0;j<SDLK_LAST;j++)
				{
					if (!strcmp(str,SDL_GetKeyName((SDLKey)j) ))
					{
					keys[i].sdl = (SDLKey)j;
					fine=true;
					}
				}
				if (!fine)
				{
				printf("Invalid keyname : '%s'\n",str);
				return -1;
				}
/* XXX check if another key uses this keybinding */
			}
		}
	}
	fclose(file);
	return 1;
}

/*!
 * \return Returns -1 if an error occurred, 0 otherwise
 * \sa loadConfig()
 */
int ControlsKeyboard::saveConfig (char *filename)
{
	int i;
	FILE *file;

	chCfgDir();
	file = fopen(filename, "w+");
	if (!file)
		return -1;

	for (i=0; i<NKEYS; i++)
	{
		fprintf(file, "%s %d\n",
			configured_keys[i].name,
			configured_keys[i].sdl);
	}

	fclose(file);
	return 0;
}

void ControlsKeyboard::strafe (float dir, Movement *movement) 
{
#if 0
	GL3f _posCam;
	if (!NETWORK)
		_posCam=stg->getCamaraPos();


	_posCam.x+=(float)cos(yaw+(M_PI/2))*speed*dir;
	_posCam.z+=(float)sin(yaw+(M_PI/2))*speed*dir;

	stg->setCamaraObj(yaw,yawV);
	if (!NETWORK)
		stg->setCamaraPos(_posCam);
#endif

	if (dir>0) setAnimation(STRAFING_RIGHT,2);
	else setAnimation(STRAFING_LEFT,2);

	float yaw = cam->getRotation();

	/* TODO use running for *2 speed */
	movement->pos.x+=cos(yaw+(M_PI/2))*speed*dir;
	movement->pos.z+=sin(yaw+(M_PI/2))*speed*dir;
}

/* TODO: Take this code to Player::setAnimation(...) (a new function?)
** TODO: Do not call this from here, instead call it from the Player::addMovement? (maybe not)
** XXX: It seems that this only works when a printf() is called O_o */
void ControlsKeyboard::setAnimation(int which,int part)
{
	int 
	ani1=(animation&3),
	ani2=((animation&28)>>2),
	ani3=((animation&224)>>5);

	//moving=true;

	switch(part)
	{
	case 1: /* body */
		switch(which)
		{
		case SHOOTING:
			switch(weaponStyle)
			{
			case FACE2FACE:
				ani2=SHOOTING_FACE2FACE;
				break;
			case _1HAND:
				ani2=SHOOTING_1HAND;
				break;
			case _2HAND:
				ani2=SHOOTING_2HAND;
				break;
			}
			break;
		case NOT_SHOOTING:
			switch(weaponStyle)
			{
			case FACE2FACE:
				ani2=NOT_SHOOTING_FACE2FACE;
				break;
			case _1HAND:
				ani2=NOT_SHOOTING_1HAND;
				break;
			case _2HAND:
				ani2=NOT_SHOOTING_2HAND;
				break;
			}
			break;
		}
		break;
	case 2: /* legs */
		switch(which)
		{
		case RUNNING_FORWARD:
		case RUNNING_BACKWARD:
		case WALKING_FORWARD:
		case WALKING_BACKWARD:
			ani3=which;
			stg->getMap()->playSound(stg->getCamaraId(),RUNNING_FORWARD,stg->getCamaraPos());
			break;
		case NOT_MOVING:
			ani3=NOT_MOVING;
			stg->getMap()->stopSound();
			break;
		default:
			ani3=which;
			stg->getMap()->playSound(stg->getCamaraId(),WALKING_FORWARD,stg->getCamaraPos());
			break;
		}
		break;
	}

	animation=0;
	animation+=(ani1);
	animation+=(ani2<<2);
	animation+=(ani3<<5);

	stg->getMyBody()->setAnimation(ani1,ani2,ani3);
}

